const __DEV__ = false
var global = window = this

this._successCallbacks = {};
this._failureCallbacks = {};
this._callID = 0;

function genModule(config, moduleID) {
  if (!config) {
    return null;
  }

  const [moduleName, constants, methods, promiseMethods, syncMethods] = config;

  if (!constants && !methods) {
    // Module contents will be filled in lazily later
    return {
      name: moduleName
    };
  }

  const module = {};
  methods && methods.forEach((methodName, methodID) => {
    const isPromise = promiseMethods && arrayContains(promiseMethods, methodID);
    const isSync = syncMethods && arrayContains(syncMethods, methodID);
    const methodType = isPromise ? 'promise' : isSync ? 'sync' : 'async';
    module[methodName] = genMethod(moduleID, methodID, methodType);
  });
  Object.assign(module, constants);

  if (module.getConstants == null) {
    module.getConstants = () => constants || Object.freeze({});
  } else {
    // console.warn(`Unable to define method 'getConstants()' on NativeModule '${moduleName}'. NativeModule '${moduleName}' already has a constant or method called 'getConstants'. Please remove it.`);
  }

  return {
    name: moduleName,
    module
  };
} // export this method as a global so we can call it from native


global.__fbGenNativeModule = genModule;


function genMethod(moduleID, methodID, type) {
  let fn = null;

  if (type === 'promise') {
    fn = function (...args) {
      return new Promise((resolve, reject) => {
        enqueueNativeCall(moduleID, methodID, args, data => resolve(data), errorData => reject(createErrorFromErrorData(errorData)));
      });
    };
  } else {
    fn = function (...args) {
      const lastArg = args.length > 0 ? args[args.length - 1] : null;
      const secondLastArg = args.length > 1 ? args[args.length - 2] : null;
      const hasSuccessCallback = typeof lastArg === 'function';
      const hasErrorCallback = typeof secondLastArg === 'function';
      const onSuccess = hasSuccessCallback ? lastArg : null;
      const onFail = hasErrorCallback ? secondLastArg : null;
      const callbackCount = hasSuccessCallback + hasErrorCallback;
      args = args.slice(0, args.length - callbackCount);

      if (type === 'sync') {
        return callNativeSyncHook(moduleID, methodID, args, onFail, onSuccess);
      } else {
        enqueueNativeCall(moduleID, methodID, args, onFail, onSuccess);
      }
    };
  }

  fn.type = type;
  return fn;
}


function arrayContains(array, value) {
  return array.indexOf(value) !== -1;
}

function createErrorFromErrorData(errorData) {
  const { message, ...extraErrorInfo } = errorData || {};
  return Object.assign({}, extraErrorInfo);
}

function callNativeSyncHook(moduleID, methodID, params, onFail, onSucc) {
  this.processCallbacks(moduleID, methodID, params, onFail, onSucc);
  return global.nativeCallSyncHook(moduleID, methodID, params);
}

function processCallbacks(moduleID, methodID, params, onFail, onSucc) {
  if (onFail || onSucc) {
    onFail && params.push(this._callID << 1); // eslint-disable-next-line no-bitwise

    onSucc && params.push(this._callID << 1 | 1);
    this._successCallbacks[this._callID] = onSucc;
    this._failureCallbacks[this._callID] = onFail;
  }
  this._callID++;
}

function enqueueNativeCall(moduleID, methodID, params, onFail, onSucc) {
  this.processCallbacks(moduleID, methodID, params, onFail, onSucc);
  global.nativeCallASyncHook(moduleID, methodID, params);
}

function invokeCallbackAndReturnFlushedQueue(cbID, args) {
  this.__guard(() => {
    this.__invokeCallback(cbID, args);
  });
}

global.__invokeCallbackAndReturnFlushedQueue = invokeCallbackAndReturnFlushedQueue;

/**
 * Private methods
 */

function __guard(fn) {
  try {
    fn();
  } catch (error) {
    // ErrorUtils.reportFatalError(error);
  }
}

function __invokeCallback(cbID, args) {
  const callID = cbID >>> 1; // eslint-disable-next-line no-bitwise

  const isSuccess = cbID & 1;
  const callback = isSuccess ? this._successCallbacks[callID] : this._failureCallbacks[callID];

  if (!callback) {
    return;
  }

  delete this._successCallbacks[callID];
  delete this._failureCallbacks[callID];
  callback(...args);
}

// 定义 NativeModules
let NativeModules = {};
if (global.nativeModuleProxy) {
  NativeModules = global.nativeModuleProxy;
}
